/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  This program is free software; you can redistribute it and/or modify    |
 |  it under the terms of the GNU General Public License as published by    |
 |  the Free Software Foundation; either version 2, or (at your option)     |
 |  any later version.                                                      |
 |                                                                          |
 |  This program is distributed in the hope that it will be useful,         |
 |  but WITHOUT ANY WARRANTY; without even the implied warranty of          |
 |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           |
 |  GNU General Public License for more details.                            |
 |                                                                          |
 |  You should have received a copy of the GNU General Public License       |
 |  along with this program; if not, write to the Free Software             |
 |  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               |
 |                                                                          |
 |  Copyright (C) 2008                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
\*--------------------------------------------------------------------------*/

///
/// file: Lapack.hh
///

#ifndef LAPACK_HH
#define LAPACK_HH

#ifndef F77NAME
  #define F77NAME(A) A##_
#endif

#ifndef ATLASNAME
  #define ATLASNAME(A) ATL_##A
#endif

namespace lapack {

  typedef char   character ;
  typedef int    integer ;
  typedef float  single_precision ;
  typedef double double_precision ;

  #ifdef USE_ATLAS

  /*
  //     #    ####### #          #     #####  
  //    # #      #    #         # #   #     # 
  //   #   #     #    #        #   #  #       
  //  #     #    #    #       #     #  #####  
  //  #######    #    #       #######       # 
  //  #     #    #    #       #     # #     # 
  //  #     #    #    ####### #     #  #####
  */

  enum ATLAS_ORDER {AtlasRowMajor=101, AtlasColMajor=102 };
  enum ATLAS_TRANS {AtlasNoTrans=111, AtlasTrans=112,
                    AtlasConjTrans=113, AtlasConj=114};
  enum ATLAS_UPLO  {AtlasUpper=121, AtlasLower=122};
  enum ATLAS_DIAG  {AtlasNonUnit=131, AtlasUnit=132};

  extern "C" {
    void
    ATLASNAME(sgetrs)( const enum ATLAS_ORDER Order,
                       const enum ATLAS_TRANS Trans,
                       const int N,
                       const int NRHS,
                       const float *A, const int lda,
                       const int *ipiv,
                       float *B, const int ldb );
    void
    ATLASNAME(dgetrs)( const enum ATLAS_ORDER Order,
                       const enum ATLAS_TRANS Trans,
                       const int N,
                       const int NRHS,
                       const double *A, const int lda,
                       const int *ipiv,
                       double *B, const int ldb );
    int
    ATLASNAME(sgetrf)( const enum ATLAS_ORDER Order,
                       const int M, const int N,
                       float *A, const int lda,
                       int *ipiv );
    int
    ATLASNAME(dgetrf)( const enum ATLAS_ORDER Order,
                       const int M, const int N,
                       double *A, const int lda,
                       int *ipiv );
    void
    ATLASNAME(strsv)( const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int N,
                      const float *A, const int lda,
                      float *X, const int incX );
    void
    ATLASNAME(dtrsv)( const enum ATLAS_UPLO  Uplo,
                      const enum ATLAS_TRANS TransA,
                      const enum ATLAS_DIAG  Diag,
                      const int N,
                      const double *A, const int lda,
                      double *X, const int incX );
    void
    ATLASNAME(szero)( const int N, float *X, const int incX );
    void
    ATLASNAME(dzero)( const int N, double *X, const int incX );

    void
    ATLASNAME(sgezero)( const int M, const int N, float *C, const int ldc );    
    void
    ATLASNAME(dgezero)( const int M, const int N, double *C, const int ldc );

    void
    ATLASNAME(sgecopy)( const int M, const int N,
                        const float *A, const int lda,
                        float       *C, const int ldc );
    void
    ATLASNAME(dgecopy)( const int M, const int N,
                        const double *A, const int lda,
                        double       *C, const int ldc );
  }

  inline
  ATLAS_TRANS
  trans_atlas(character const TRANS ) {
    switch ( TRANS ) {
    case 'N': return AtlasNoTrans ;
    case 'T': return AtlasTrans ;
    case 'C': return AtlasConjTrans ;
    default:  throw runtime_error("bad traspose mode selection\n") ;
    }
  }

  inline
  ATLAS_UPLO
  uplo_atlas(character const UPLO ) {
    switch ( UPLO ) {
    case 'U': return AtlasUpper ;
    case 'L': return AtlasLower ;
    default:  throw runtime_error("bad UP/LOW mode selection\n") ;
    }
  }

  inline
  ATLAS_DIAG
  diag_atlas(character const DIAG ) {
    switch ( DIAG ) {
    case 'U': return AtlasUnit ;
    case 'N': return AtlasNonUnit ;
    default:  throw runtime_error("bad diagonal mode selection\n") ;
    }
  }

  #else

  /*
  //  #        ##   #####    ##    ####  #    # 
  //  #       #  #  #    #  #  #  #    # #   #  
  //  #      #    # #    # #    # #      ####   
  //  #      ###### #####  ###### #      #  #   
  //  #      #    # #      #    # #    # #   #  
  //  ###### #    # #      #    #  ####  #    # 
  */

  extern "C" {
    void
    F77NAME(dgetrf)( integer          const * N, 
                     integer          const * M, 
                     double_precision         A[], 
                     integer          const * LDA, 
                     integer                  IPIV[], 
                     integer                * INFO ) ;
    void
    F77NAME(sgetrs)( character        const   TRANS[], 
                     integer          const * N, 
                     integer          const * NRHS, 
                     single_precision const   A[], 
                     integer          const * LDA, 
                     integer          const   IPIV[], 
                     single_precision         B[], 
                     integer          const * LDB, 
                     integer                * INFO ) ;
    void
    F77NAME(dgetrs)( character        const   TRANS[],
                     integer          const * N, 
                     integer          const * NRHS, 
                     double_precision const   A[], 
                     integer          const * LDA, 
                     integer          const   IPIV[], 
                     double_precision         B[], 
                     integer          const * LDB, 
                     integer                * INFO ) ;
    void
    F77NAME(sgetrf)( integer          const * N, 
                     integer          const * M, 
                     single_precision         A[], 
                     integer          const * LDA, 
                     integer                  IPIV[], 
                     integer                * INFO ) ;
    void
    F77NAME(dgetrf)( integer          const * N, 
                     integer          const * M, 
                     double_precision         A[], 
                     integer          const * LDA, 
                     integer                  IPIV[], 
                     integer                * INFO ) ;
    void
    F77NAME(sgetrs)( character        const   TRANS[], 
                     integer          const * N, 
                     integer          const * NRHS, 
                     single_precision const   A[], 
                     integer          const * LDA, 
                     integer          const   IPIV[], 
                     single_precision         B[], 
                     integer          const * LDB, 
                     integer                * INFO ) ;
    void
    F77NAME(dgetrs)( character        const   TRANS[],
                     integer          const * N, 
                     integer          const * NRHS, 
                     double_precision const   A[], 
                     integer          const * LDA, 
                     integer          const   IPIV[], 
                     double_precision         B[], 
                     integer          const * LDB, 
                     integer                * INFO ) ;
    void 
    F77NAME(sgesv)( integer          const * N, 
                    integer          const * NRHS, 
                    single_precision         A[], 
                    integer          const * LDA, 
                    integer                  IPIV[], 
                    single_precision         B[], 
                    integer          const * LDB, 
                    integer                * INFO ) ;
    void
    F77NAME(dgesv)( integer          const * N, 
                    integer          const * NRHS, 
                    double_precision         A[], 
                    integer          const * LDA, 
                    integer                  IPIV[], 
                    double_precision         B[], 
                    integer          const * LDB, 
                    integer                * INFO ) ;
    void
    F77NAME(strsv)( character        const   UPLO[], 
                    character        const   TRANS[], 
                    character        const   DIAG[], 
                    integer          const * N, 
                    single_precision const   A[], 
                    integer          const * LDA, 
                    single_precision         X[], 
                    integer          const * INCX ) ;
    void
    F77NAME(dtrsv)( character        const   UPLO[], 
                    character        const   TRANS[], 
                    character        const   DIAG[], 
                    integer          const * N, 
                    double_precision const   A[], 
                    integer          const * LDA, 
                    double_precision         X[], 
                    integer          const * INCX ) ;
    void
    F77NAME(slaset)( character        const   UPLO[], 
                     integer          const * M, 
                     integer          const * N, 
                     single_precision const * ALPHA, 
                     single_precision const * BETA, 
                     single_precision         A[], 
                     integer          const * LDA ) ;
    void
    F77NAME(dlaset)( character        const   UPLO[], 
                     integer          const * M, 
                     integer          const * N, 
                     double_precision const * ALPHA, 
                     double_precision const * BETA, 
                     double_precision         A[], 
                     integer          const * LDA ) ;
    void
    F77NAME(slacpy)( character        const   UPLO[], 
                     integer          const * M, 
                     integer          const * N, 
                     single_precision const   A[], 
                     integer          const * LDA, 
                     single_precision         B[], 
                     integer          const * LDB) ;
    void
    F77NAME(dlacpy)( character        const   UPLO[], 
                     integer          const * M, 
                     integer          const * N, 
                     double_precision const   A[], 
                     integer          const * LDA, 
                     double_precision         B[], 
                     integer          const * LDB ) ;
  }

  #endif

  /* ____ ____ ___  _   _ 
  // |    |  | |__]  \_/  
  // |___ |__| |      |   
  */
  inline
  void
  gecopy( integer          const M, 
          integer          const N, 
          single_precision const A[], 
          integer          const LDA, 
          single_precision       B[], 
          integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(sgecopy)( M, N, A, LDA, B, LDB ) ; }
  #else
  { F77NAME(slacpy)( "A", &M, &N, A, &LDA, B, &LDB ) ; }
  #endif

  inline
  void
  gecopy( integer          const M, 
          integer          const N, 
          double_precision const A[], 
          integer          const LDA, 
          double_precision       B[], 
          integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(dgecopy)( M, N, A, LDA, B, LDB ) ; }
  #else
  { F77NAME(dlacpy)( "A", &M, &N, A, &LDA, B, &LDB ) ; }
  #endif

  /*  __  _  _   _  
  //   / |_ |_) / \ 
  //  /_ |_ | \ \_/
  */
  inline
  void
  gezero( integer          const M, 
          integer          const N, 
          single_precision       A[], 
          integer          const LDA )
  #ifdef USE_ATLAS
  { ATLASNAME(sgezero)( M, N, A, LDA ) ; }
  #else
  { single_precision zero = 0 ;
    F77NAME(slaset)( "A", &M, &N, &zero, &zero, A, &LDA ) ; }
  #endif

  inline
  void
  gezero( integer          const M, 
          integer          const N, 
          double_precision       A[], 
          integer          const LDA )
  #ifdef USE_ATLAS
  { ATLASNAME(dgezero)( M, N, A, LDA ) ; }
  #else
  { double_precision zero = 0 ;
    F77NAME(dlaset)( "A", &M, &N, &zero, &zero, A, &LDA ) ; }
  #endif

  /*   ___ _   __     
  //    | |_) (_ \  / 
  //    | | \ __) \/  
  */
  inline
  void
  trsv( character        const UPLO[], 
        character        const TRANS[], 
        character        const DIAG[],
        integer          const N, 
        single_precision const A[], 
        integer          const LDA, 
        single_precision       X[], 
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(strsv)( uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      N, A, LDA, X, INCX ) ; }
  #else
  { F77NAME(strsv)( UPLO, TRANS, DIAG, &N, A, &LDA, X, &INCX ) ; }
  #endif

  inline
  void
  trsv( character        const UPLO[],
        character        const TRANS[],
        character        const DIAG[],
        integer          const N, 
        double_precision const A[], 
        integer          const LDA, 
        double_precision       X[], 
        integer          const INCX )
  #ifdef USE_ATLAS
  { ATLASNAME(dtrsv)( uplo_atlas(UPLO[0]),
                      trans_atlas(TRANS[0]),
                      diag_atlas(DIAG[0]),
                      N, A, LDA, X, INCX ) ; }
  #else
  { F77NAME(dtrsv)( UPLO, TRANS, DIAG, &N, A, &LDA, X, &INCX ) ; }
  #endif

  /*   __  _ ___ _   _ 
  //  /__ |_  | |_) |_ 
  //  \_| |_  | | \ |  
  */
  inline
  integer
  getrf( integer          const N, 
         integer          const M, 
         single_precision       A[], 
         integer          const LDA, 
         integer                IPIV[] )
  #ifdef USE_ATLAS
  { return ATLASNAME(sgetrf)( AtlasColMajor, N, M, A, LDA, IPIV ) ; }
  #else
  { integer INFO ;
    F77NAME(sgetrf)( &N, &M, A, &LDA, IPIV, &INFO ) ;
    return INFO ; }
  #endif

  inline
  integer
  getrf( integer          const N,
         integer          const M,
         double_precision       A[],
         integer          const LDA, 
         integer                IPIV[] )
  #ifdef USE_ATLAS
  { return ATLASNAME(dgetrf)( AtlasColMajor, N, M, A, LDA, IPIV ) ; }
  #else
  { integer INFO ;
    F77NAME(dgetrf)( &N, &M, A, &LDA, IPIV, &INFO ) ;
    return INFO ; }
  #endif

  /*   __  _ ___ _   __ 
  //  /__ |_  | |_) (_  
  //  \_| |_  | | \ __) 
  */
  inline
  integer
  getrs( character        const TRANS[], 
         integer          const N, 
         integer          const NRHS, 
         single_precision const A[], 
         integer          const LDA, 
         integer          const IPIV[], 
         single_precision       B[], 
         integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(sgetrs)( AtlasColMajor, trans_atlas(TRANS[0]), N, NRHS, A, LDA, IPIV, B, LDB ) ;
    return 0 ; }
  #else
  { integer INFO ; 
    F77NAME(sgetrs)( TRANS, &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  inline
  integer
  getrs( character        const TRANS[], 
         integer          const N, 
         integer          const NRHS, 
         double_precision const A[], 
         integer          const LDA, 
         integer          const IPIV[], 
         double_precision       B[], 
         integer          const LDB )
  #ifdef USE_ATLAS
  { ATLASNAME(dgetrs)( AtlasColMajor, trans_atlas(TRANS[0]), N, NRHS, A, LDA, IPIV, B, LDB ) ;
    return 0 ; }
  #else
  { integer INFO ;
    F77NAME(dgetrs)( TRANS, &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  /*   __  _  __
  //  /__ |_ (_ \  / 
  //  \_| |_ __) \/  
  */
  inline
  integer
  gesv( integer          const N, 
        integer          const NRHS, 
        single_precision       A[], 
        integer          const LDA, 
        integer                IPIV[], 
        single_precision       B[], 
        integer          const LDB )
  #ifdef USE_ATLAS
  { integer INFO = ATLASNAME(sgetrf)(AtlasColMajor, N, N, A, LDA, IPIV);
    if ( INFO == 0 ) ATLASNAME(sgetrs)(AtlasColMajor, AtlasNoTrans,
                                       N, NRHS, A, LDA, IPIV, B, LDB) ;
    return INFO ; }
  #else
  { integer INFO ;
    F77NAME(sgesv)( &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  inline
  integer
  gesv( integer          const N, 
        integer          const NRHS, 
        double_precision       A[], 
        integer          const LDA, 
        integer                IPIV[], 
        double_precision       B[], 
        integer          const LDB )
  #ifdef USE_ATLAS
  { integer INFO = ATLASNAME(dgetrf)(AtlasColMajor, N, N, A, LDA, IPIV);
    if ( INFO == 0 ) ATLASNAME(dgetrs)(AtlasColMajor, AtlasNoTrans,
                                       N, NRHS, A, LDA, IPIV, B, LDB) ;
    return INFO ; }
  #else
  { integer INFO ;
    F77NAME(dgesv)( &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO ) ;
    return INFO ; }
  #endif

  extern "C" {

    void
    F77NAME(sgttrf)( integer          const * N,
                     single_precision         DL[], 
                     single_precision         D[], 
                     single_precision         DU[], 
                     single_precision         DU2[], 
                     integer                  IPIV[], 
                     integer                * INFO ) ;

    void
    F77NAME(dgttrf)( integer          const * N,
                     double_precision         DL[], 
                     double_precision         D[], 
                     double_precision         DU[], 
                     double_precision         DU2[], 
                     integer                  IPIV[], 
                     integer                * INFO ) ;

    void
    F77NAME(sgttrs)( character                TRANS[],
                     integer          const * N,
                     integer          const * NRHS, 
                     single_precision const   DL[], 
                     single_precision const   D[], 
                     single_precision const   DU[], 
                     single_precision const   DU2[], 
                     integer          const   IPIV[], 
                     single_precision         B[], 
                     integer          const * LDB, 
                     integer                * INFO ) ;

    void
    F77NAME(dgttrs)( character                TRANS[],
                     integer          const * N,
                     integer          const * NRHS, 
                     double_precision const   DL[], 
                     double_precision const   D[], 
                     double_precision const   DU[], 
                     double_precision const   DU2[], 
                     integer          const   IPIV[], 
                     double_precision         B[], 
                     integer          const * LDB, 
                     integer                * INFO ) ;

    void
    F77NAME(sgtsv)( integer          const * N, 
                    integer          const * NRHS, 
                    single_precision         DL[], 
                    single_precision         D[], 
                    single_precision         DU[], 
                    single_precision         B[],
                    integer          const * LDB, 
                    integer                * INFO ) ;

    void 
    F77NAME(dgtsv)( integer          const * N, 
                    integer          const * NRHS, 
                    double_precision         DL[], 
                    double_precision         D[], 
                    double_precision         DU[], 
                    double_precision         B[], 
                    integer          const * LDB, 
                    integer                * INFO ) ;
  }

  /*   __ ___ ___ _   _ 
  //  /__  |   | |_) |_ 
  //  \_|  |   | | \ |  
  */

  inline
  integer
  gttrf( integer  const   N,
         single_precision DL[], 
         single_precision D[], 
         single_precision DU[], 
         single_precision DU2[], 
         integer          IPIV[] )
  { integer INFO ;
    F77NAME(sgttrf)( &N, DL, D, DU, DU2, IPIV, &INFO) ;
    return INFO ;  }

  inline
  integer
  gttrf( integer  const   N,
         double_precision DL[], 
         double_precision D[], 
         double_precision DU[], 
         double_precision DU2[], 
         integer          IPIV[] )
  { integer INFO ;
    F77NAME(dgttrf)( &N, DL, D, DU, DU2, IPIV, &INFO) ;
    return INFO ; }

  /*   __ ___ ___ _   __ 
  //  /__  |   | |_) (_  
  //  \_|  |   | | \ __) 
  */

  inline
  integer
  gttrs( character              TRANS[],
         integer          const N,
         integer          const NRHS, 
         single_precision const DL[], 
         single_precision const D[], 
         single_precision const DU[], 
         single_precision const DU2[], 
         integer          const IPIV[], 
         single_precision       B[], 
         integer          const LDB )
  { integer INFO ;
    F77NAME(sgttrs)( TRANS, &N, &NRHS, DL, D, DU, DU2, IPIV, B, &LDB, &INFO) ;
    return INFO ; }

  inline
  integer
  gttrs( character              TRANS[],
         integer          const N,
         integer          const NRHS, 
         double_precision const DL[], 
         double_precision const D[], 
         double_precision const DU[], 
         double_precision const DU2[], 
         integer          const IPIV[], 
         double_precision       B[], 
         integer          const LDB )
  { integer INFO ;
    F77NAME(dgttrs)( TRANS, &N, &NRHS, DL, D, DU, DU2, IPIV, B, &LDB, &INFO) ;
    return INFO ; }

  /*   __ ___ __     
  //  /__  | (_ \  / 
  //  \_|  | __) \/  
  */

  inline
  integer
  gtsv( integer  const   N,
        integer  const   NRHS,
        single_precision DL[],
        single_precision D[],
        single_precision DU[],
        single_precision B[], 
        integer  const   LDB )
  { integer INFO ;
    F77NAME(sgtsv)( &N, &NRHS, DL, D, DU, B, &LDB, &INFO ) ;
    return INFO ; }

  inline
  integer
  gtsv( integer  const   N,
        integer  const   NRHS,
        double_precision DL[],
        double_precision D[],
        double_precision DU[],
        double_precision B[], 
        integer  const   LDB )
  { integer INFO ; 
    F77NAME(dgtsv)( &N, &NRHS, DL, D, DU, B, &LDB, &INFO ) ;
    return INFO ; }

} // end namespace lapack

#endif

///
/// eof: Lapack.hh
///

